import { createToken, createPlugin } from 'fusion-core';
import { UniversalEventsToken } from 'fusion-plugin-universal-events';
import { I18nToken } from 'fusion-plugin-i18n';
import { FetchToken } from 'fusion-tokens';
import mapObject from 'just-map-object';
import isEqual from 'just-compare';

/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
const RPCToken = createToken('RPCToken');
const RPCHandlersToken = createToken('RPCHandlersToken');
const BodyParserOptionsToken = createToken('BodyParserOptionsToken');
const RPCHandlersConfigToken = createToken('RPCHandlersConfigToken');

const formatApiPath = apiPath => `/${apiPath}/`.replace(/\/{2,}/g, '/');

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(source, true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(source).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */

/* eslint-env browser */
const statKey = 'rpc:method-client';

class RPC {
  constructor({
    fetch,
    emitter,
    rpcConfig,
    localeCode
  }) {
    this.fetch = fetch;
    this.config = rpcConfig || {};
    this.emitter = emitter;
    this.localeCode = localeCode;
    this.apiPath = formatApiPath(rpcConfig && rpcConfig.apiPath ? rpcConfig.apiPath : 'api');
  }

  request(rpcId, args, headers) {
    if (!this.fetch) {
      throw new Error('fusion-plugin-rpc requires `fetch`');
    }

    if (!this.emitter) {
      throw new Error('Missing emitter registered to UniversalEventsToken');
    }

    const fetch = this.fetch;
    const emitter = this.emitter;
    const apiPath = this.apiPath;
    const startTime = Date.now();
    const localeParam = this.localeCode ? `?localeCode=${this.localeCode}` : '';
    return fetch(`${apiPath}${rpcId}${localeParam}`, args instanceof FormData ? {
      method: 'POST',
      headers: _objectSpread({}, headers || {}),
      body: args
    } : {
      method: 'POST',
      headers: _objectSpread({
        'Content-Type': 'application/json'
      }, headers || {}),
      body: JSON.stringify(args || {})
    }).then(r => r.json()).then(args => {
      const {
        status,
        data
      } = args;

      if (status === 'success') {
        emitter.emit(statKey, {
          method: rpcId,
          status: 'success',
          timing: Date.now() - startTime
        });
        return data;
      } else {
        emitter.emit(statKey, {
          method: rpcId,
          error: data,
          status: 'failure',
          timing: Date.now() - startTime
        });
        return Promise.reject(data);
      }
    });
  }

}

const pluginFactory = () => createPlugin({
  deps: {
    fetch: FetchToken,
    emitter: UniversalEventsToken,
    i18n: I18nToken.optional,
    rpcConfig: RPCHandlersConfigToken.optional
  },
  provides: deps => {
    const {
      fetch = window.fetch,
      emitter,
      rpcConfig,
      i18n
    } = deps;
    return {
      from: ctx => {
        const locale = i18n && i18n.from(ctx).locale || '';
        const localeCode = typeof locale === 'string' ? locale : locale.code;
        return new RPC({
          fetch,
          emitter,
          rpcConfig,
          localeCode
        });
      }
    };
  }
});

var browserDataFetching = true && pluginFactory();

/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
class MissingHandlerError extends Error {
  constructor(method) {
    super(`Missing RPC handler for ${method}`);
    this.code = 'ERR_MISSING_HANDLER';
    Error.captureStackTrace(this, MissingHandlerError);
  }

}

/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
class ResponseError extends Error {
  constructor(message) {
    super(message);
    this.code = null;
    this.meta = null;
    Error.captureStackTrace(this, ResponseError);
  }

}

/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */

/* eslint-env node */

/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
class RPC$2 {
  constructor(handlers) {
    this.handlers = handlers;
  }

  async request(method, args) {
    if (!this.handlers) {
      throw new Error('fusion-plugin-rpc requires `handlers`');
    }

    if (!this.handlers[method]) {
      throw new MissingHandlerError(method);
    }

    return this.handlers[method](args);
  }

}

const plugin = createPlugin({
  deps: {
    handlers: RPCHandlersToken,
    // $FlowFixMe
    emitter: UniversalEventsToken.optional
  },
  provides: ({
    handlers
  } = {}) => {
    return {
      from: () => new RPC$2(handlers)
    };
  }
});

function ownKeys$1(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$1(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$1(source, true).forEach(function (key) { _defineProperty$1(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$1(source).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty$1(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const getMockRpcHandlers = (fixtures, onMockRpc) => fixtures.reduce((rpcHandlers, fixture) => _objectSpread$1({}, rpcHandlers, {}, mapObject(fixture, (rpcId, responseDetails) => async (...args) => {
  const response = Array.isArray(responseDetails) ? responseDetails.filter(item => isEqual(item.args, args))[0].response : responseDetails;
  onMockRpc && onMockRpc(rpcId, args, response);

  if (response instanceof Error) {
    throw response;
  }

  return response;
})), {});

/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
var index = browserDataFetching;

export default index;
export { plugin as mock, ResponseError, BodyParserOptionsToken, RPCToken, RPCHandlersToken, RPCHandlersConfigToken, getMockRpcHandlers };
